# Copyright (c) 2016 FUJITSU LIMITED
# Copyright (c) 2012 EMC Corporation.
# Copyright (c) 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""
Cinder Volume driver for Fujitsu ETERNUS DX S3 series.

This driver is extension version.
"""

# $Date:: 2016-03-10 16:51:26 +0800#$
# $Revision: 337 $

import codecs
import os
import threading
import time
import uuid
from xml.dom.minidom import Document
from xml.dom.minidom import parse as xml_parse
from xml.etree.ElementTree import parse as tree_parse

from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_config import types
from oslo_log import log as logging
import six

from cinder import context
from cinder import db
from cinder import exception
from cinder import utils
from cinder.i18n import _, _LE, _LI, _LW
from cinder.volume.drivers.fujitsu import eternus_dx_common
from cinder.volume.drivers.fujitsu import eternus_dx_utils as dx_utils
from cinder.volume import qos_specs
from cinder.volume import utils as volume_utils
from cinder.volume import volume_types

LOG = logging.getLogger(__name__)

try:
    import pywbem
except ImportError:
    msg = _LE('import pywbem failed!! '
              'pywbem is necessary for this volume driver.')
    LOG.error(msg)

CONF = cfg.CONF
MONITOR_IMGVOL_INTERVAL = 600
DELETE_IMGVOL = "Deleting"
FJ_REMOTE_SRC_META = "FJ_Remote_Copy_Source"
FJ_REMOTE_DEST_META = "FJ_Remote_Copy_Destination"
FJ_REC_CLONE = "clone"
FJ_REC_MIRROR = "mirror"
FJ_VOL_FORMAT_KEY = "type:delete_with_volume_format"
LIMIT_SESSION_PER_VOLUME = 8  # limitation of OPC session by volume
FJ_SRC_VOL_ID = "FJ_Source_Volume_ID"
EC_REC = 3
RAIDGROUP = 2
TPPOOL = 5
CTRL_CONF = "FUJITSU_ControllerConfigurationService"

FJ_ETERNUS_DX_OPT_list = [
    cfg.StrOpt('fujitsu_min_image_volume_per_storage',
               default='0',
               help='minimum number of image volume '
               'per storage(ETERNUS)'),
]

CINDER_CONF_OPT_opts = [
    cfg.StrOpt('fujitsu_image_management_dir',
               default=CONF.image_conversion_dir,
               help='directory for '
               'image management file(xml)'),
    cfg.MultiOpt('replication_device',
                 item_type=types.Dict(),
                 help='Multi opt of dictionaries to represent a replication '
                 'target device. We only support one replication '
                 'target device. The replication target device takes the '
                 'standard dict config form: replication_device = '
                 'managed_backend_name:<ctrl.domain@DX_Remote#dx_R>, '
                 'remote_eternus_config:<dx_R.xml>'),
]

FJ_QOS_KEY_list = ['maxBWS']

SPECIAL_STATS_dic = {'QoS_support': True}

POOL_TYPE_dic = {
    RAIDGROUP: 'RAID_GROUP',
    TPPOOL: 'Thinporvisioning_POOL',
}

RETCODE_dic = {
    '32812': 'This operation cannot be performed to the '
             'Transparent Failover resources',
}

CONF.register_opts(CINDER_CONF_OPT_opts)


class FJDXCommonExt(eternus_dx_common.FJDXCommon):
    """Common code that does not depend on protocol."""

    def __init__(self, prtcl, configuration=None):
        self.supercls = super(FJDXCommonExt, self)
        self.supercls.__init__(prtcl, configuration)
        self.stats.update(SPECIAL_STATS_dic)
        self.configuration.append_config_values(FJ_ETERNUS_DX_OPT_list)
        self.configuration.fujitsu_image_management_file = (
            CONF.fujitsu_image_management_dir + (
                '/' + 'cinder_fujitsu_image_management.xml'))
        imag_vol_num = self.configuration.fujitsu_min_image_volume_per_storage
        use_fj_imag_vol = self.configuration.use_fujitsu_image_volume
        if ((imag_vol_num.isdigit() is False) or (use_fj_imag_vol is False)):
            self.configuration.fujitsu_min_image_volume_per_storage = '0'

        self._check_user()
        self.invalid_migration_list = []

    def create_volume(self, volume):
        """Create volume on ETERNUS."""
        LOG.debug('create_volume, '
                  'volume id: %(vid)s, volume size: %(vsize)s.',
                  {'vid': volume['id'], 'vsize': volume['size']})

        model_update = self.supercls.create_volume(volume)

        # Set qos to created volume.
        self._set_qos(volume, use_id=True)

        repl_value = dx_utils.get_extra_specs(
            volume, eternus_dx_common.FJ_VOL_REPL_KEY)

        if repl_value == eternus_dx_common.FJ_VOL_REPL_VALUE:
            # Set qos to remote volume.
            self._set_qos(volume, self.remote_eternus_config)

        return model_update

    def create_volume_from_snapshot(self, volume, snapshot):
        """Create a volume from a snapshot."""
        LOG.debug('create_volume_from_snapshot, '
                  'volume id: %(vid)s, volume size: %(vsize)s, '
                  'snapshot id: %(sid)s.',
                  {'vid': volume['id'], 'vsize': volume['size'],
                   'sid': snapshot['id']})

        self.conn = self._get_eternus_connection()
        source_volume_instance = self._find_lun(snapshot)

        # Check the existence of source volume.
        if not source_volume_instance:
            msg = _('create_volume_from_snapshot, '
                    'Source Volume is not exist in ETERNUS.')
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        # Create volume for destination of cloned volume.
        model_update = self.create_volume(volume)
        element_path = eval(model_update.get('provider_location'))
        target_volume_instancename = self._create_eternus_instance_name(
            element_path['classname'], element_path['keybindings'])
        target_volume_instance = self._get_eternus_instance(
            target_volume_instancename)
        self._create_local_cloned_volume(target_volume_instance,
                                         source_volume_instance)

        return model_update

    def create_cloned_volume(self, volume, src_vref,
                             CloneOnly=False, use_service_name=False):
        """Create clone of the specified volume."""
        LOG.debug('create_cloned_volume, '
                  'tgt: (%(tid)s), src: (%(sid)s).',
                  {'tid': volume['id'], 'sid': src_vref['id']})

        model_update = None

        self.conn = self._get_eternus_connection()
        source_volume_instance = self._find_lun(src_vref, use_service_name)

        if not source_volume_instance:
            msg = _('create_cloned_volume, '
                    'Source Volume is not exist in ETERNUS.')
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        if CloneOnly is False:
            model_update = self.create_volume(volume)
            element_path = eval(model_update.get('provider_location'))
            metadata = model_update.get('metadata')
            target_volume_instancename = self._create_eternus_instance_name(
                element_path['classname'], element_path['keybindings'])
            target_volume_instance = self._get_eternus_instance(
                target_volume_instancename)
        else:
            target_volume_instance = self._find_lun(volume)
            try:
                element_path = eval(volume['provider_location'])
                metadata = volume['volume_metadata']
            except Exception:
                element_path = None
                metadata = None

        remote_copy_type = (
            dx_utils.get_metadata(volume).get(FJ_REMOTE_DEST_META, None))
        if not remote_copy_type:
            self._create_local_cloned_volume(target_volume_instance,
                                             source_volume_instance)
        else:
            source_volume_metadata = dx_utils.get_metadata(src_vref)
            self._create_remote_cloned_volume(
                volume, metadata, src_vref,
                source_volume_metadata, remote_copy_type)

        if not model_update:
            model_update = {
                'provider_location': six.text_type(element_path),
                'metadata': metadata,
            }

        return model_update

    def _create_remote_cloned_volume(self, volume, volume_metadata, src_vref,
                                     src_vref_metadata, remote_copy_type):
        """Create remote clone of the specified volume."""
        LOG.debug('_create_remote_cloned_volume, '
                  'tgt: %(tid)s, src: %(sid)s.',
                  {'tid': volume['id'], 'sid': src_vref['id']})

        ctxt = context.get_admin_context()
        copy_type = str(remote_copy_type).lower()

        # Check if source volume has already get a mirror target.
        source_volume_copy_dict = (
            eval(src_vref_metadata.get(FJ_REMOTE_SRC_META, "{}")))

        # When remote_copy_type is mirror, check mirror copy already exist.
        if copy_type == FJ_REC_MIRROR:
            for value in source_volume_copy_dict.values():
                if value == FJ_REC_MIRROR:
                    s_volumename = self._get_volume_name(src_vref)
                    msg = (_('_create_remote_cloned_volume, '
                             'source volume: %(src)s already has a mirror!')
                           % {'src': s_volumename})
                    LOG.error(msg)
                    raise exception.VolumeBackendAPIException(data=msg)

        # Execute CCM script.
        self._exec_ccm_script("start",
                              source=src_vref_metadata,
                              source_volume=src_vref,
                              target=volume_metadata,
                              copy_type=copy_type)

        # Update metadata of source volume.
        lockname = 'db-update-' + src_vref['id']

        @dx_utils.FJDXLockutils(lockname, 'cinder-', True)
        def __update_metadata(self):
            """Update volume metadata."""
            LOG.debug('__update_metadata, '
                      'copy_type: %s.', copy_type)
            source_volume_copy_dict = {volume['id']: copy_type}
            metadata = (
                {FJ_REMOTE_SRC_META: six.text_type(source_volume_copy_dict)})
            db.volume_metadata_update(ctxt.elevated(), src_vref['id'],
                                      metadata, False)

        if copy_type == FJ_REC_MIRROR:
            # If remote copy type is mirror, update metadata.
            __update_metadata(self)
        elif copy_type == FJ_REC_CLONE:
            # If remote copy type and copy_list is {}, update metadata.
            if not source_volume_copy_dict:
                __update_metadata(self)

    def delete_volume(self, volume):
        """Delete volume on ETERNUS."""
        LOG.debug('delete_volume, volume id: %s.', volume['id'])

        deleted = False
        is_rec_copy_stopped = False

        self.conn = self._get_eternus_connection()
        vol_instance = self._find_lun(volume)
        if not vol_instance:
            return False

        # CCM processing
        metadata = dx_utils.get_metadata(volume)
        copy_type = metadata.get(FJ_REMOTE_DEST_META, None)

        if copy_type and ("FJ_Backend" in metadata):
            self._exec_ccm_script("stop", target=metadata)

            # If copy_type is mirror, update FJ_Remote_Copy_Source info.
            tmp_flag1 = FJ_SRC_VOL_ID in metadata
            tmp_flag2 = str(copy_type).lower() == FJ_REC_MIRROR
            if tmp_flag1 and tmp_flag2:
                source_volume_id = metadata.get(FJ_SRC_VOL_ID)
                ctxt = context.get_admin_context()
                src_vref = db.volume_get(ctxt, source_volume_id)
                source_volume_metadata = dx_utils.get_metadata(src_vref)
                source_volume_copy_list = (
                    eval(source_volume_metadata.get(FJ_REMOTE_SRC_META, "{}")))

                # If source_volume_copy_list has this volume, delete it.
                volume_id = volume['id']
                if volume_id in source_volume_copy_list:
                    del source_volume_copy_list[volume_id]

                update_metadata = (
                    {FJ_REMOTE_SRC_META:
                     six.text_type(source_volume_copy_list)})
                db.volume_metadata_update(ctxt.elevated(),
                                          source_volume_id,
                                          update_metadata,
                                          False)

        if metadata.get(FJ_REMOTE_SRC_META, None):
            self._exec_ccm_script(
                "stop", source=metadata, source_volume=volume)
            is_rec_copy_stopped = True

            # Update volume metadata.
            copy_list = eval(metadata.get(FJ_REMOTE_SRC_META))
            ctxt = context.get_admin_context()
            for target_volume_id, copy_type in copy_list.items():
                try:
                    if copy_type == FJ_REC_MIRROR:
                        db.volume_metadata_update(
                            ctxt.elevated(), target_volume_id,
                            {FJ_REMOTE_DEST_META: None}, False)
                except Exception:
                    pass

        with_format = dx_utils.get_extra_specs(volume,
                                               key=FJ_VOL_FORMAT_KEY,
                                               default='False')

        repl_status = volume.get('replication_status', None)
        if repl_status and repl_status != 'disabled':
            remote_vol = None
            try:
                remote_vol = self._find_replica_lun(volume)
            except Exception:
                LOG.warning(_LW('delete_volume, '
                                'remote volume does not found.'))

            if remote_vol:
                repl_deleted = False
                metadata = dx_utils.get_metadata(volume)
                if not is_rec_copy_stopped and repl_status == 'enabled':
                    self._exec_ccm_script(
                        "stop", source=metadata, source_volume=volume)

                src_host = volume_utils.extract_host(volume['host'], 'host')
                if self.remote_host != src_host:
                    msg = (_('Do replication, '
                             'remote host:%(remote)s '
                             'and local host:%(local)s '
                             'must be the same one.')
                           % {'remote': self.remote_host,
                              'local': src_host})
                    LOG.error(msg)
                    raise exception.VolumeBackendAPIException(data=msg)

                self.remote_conn = self._get_eternus_connection(
                    self.remote_eternus_config)

                repl_vol_exist = self._delete_volume_setting(volume,
                                                             remote=True)

                if not repl_vol_exist:
                    LOG.debug('delete_repl_volume, '
                              'volume not found in 1st check.')
                else:
                    try:
                        repl_deleted = self._delete_volume_base(
                            volume, with_format, remote=True)
                    except Exception:
                        msg = (_LE('delete_volume, '
                                   'delete replica volume failed, '
                                   'deleted status is: %s.')
                               % repl_deleted)
                        LOG.error(msg)
                        raise exception.VolumeBackendAPIException(data=msg)

        vol_exist = self._delete_volume_setting(volume)

        if not vol_exist:
            LOG.debug('delete_volume, volume not found in 1st check.')
            return False

        try:
            deleted = self._delete_volume_base(volume, with_format)
        except Exception:
            msg = (_LE('delete_volume, '
                       'delete volume failed, '
                       'deleted status is: %s.')
                   % deleted)
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('delete_volume, deleted status is: %s.', deleted)

        return deleted

    def _format_standard_volume(self, volume, remote=False):
        """Format standard volume."""
        LOG.debug('_format_standard_volume, volume id: %(vid)s, '
                  'remote: %(remote)s',
                  {'vid': volume['id'], 'remote': remote})

        wait_sec = 60
        wait_cnt = 0
        filename = None

        volumename = self._get_volume_name(volume)
        param_dict = {'volume-name': volumename}

        if remote:
            filename = self.remote_eternus_config

        rc, errordesc, job = self._exec_eternus_cli('format_volume',
                                                    filename=filename,
                                                    **param_dict)

        if rc != 0:
            msg = (_('_format_standard_volume, '
                     'volumename: %(volumename)s, '
                     'Return code: %(rc)lu, '
                     'Error: %(errordesc)s, '
                     'Job: %(job)s.')
                   % {'volumename': volumename,
                      'rc': rc,
                      'errordesc': errordesc,
                      'job': job})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        while True:
            time.sleep(wait_sec)
            rc, errordesc, job = self._exec_eternus_cli(
                'show_volume_progress', filename=filename, **param_dict)

            if rc != 0:
                msg = (_('_format_standard_volume, '
                         'show volume progress error, '
                         'volumename: %(volumename)s, '
                         'Return code: %(rc)lu, '
                         'Error: %(errordesc)s, '
                         'Job: %(job)s.')
                       % {'volumename': volumename,
                          'rc': rc,
                          'errordesc': errordesc,
                          'job': job})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            wait_sec = int(job)

            if wait_sec == 0:
                LOG.debug('_format_standard_volume, format complete.')
                break
            elif wait_cnt < 3:
                wait_sec /= 3
                wait_cnt += 1

            LOG.debug('_format_standard_volume, format remain time: %s sec.',
                      job)

    def _format_tpv(self, volume, vol_instance, poolname=None, remote=False):
        """Format tpv."""
        LOG.debug('_format_tpv, volume id: %(vid)s, remote: %(remote)s',
                  {'vid': volume['id'], 'remote': remote})

        filename = None
        volumename = self._get_volume_name(volume)

        if remote:
            filename = self.remote_eternus_config

        if not poolname:
            poolname, _ = self._find_pool_from_volume(vol_instance,
                                                      remote=remote)

        lockname = 'format-' + poolname + six.text_type(remote)

        @dx_utils.FJDXLockutils(lockname, 'cinder-', True)
        def __format_tpv(self, volumename, poolname, filename):
            """Do format tpv using CLI method."""
            wait_sec = 60
            wait_cnt = 0

            format_param_dict = {'volume-name': volumename}

            rc, errordesc, job = self._exec_eternus_cli('format_tpv',
                                                        filename=filename,
                                                        **format_param_dict)

            if rc != 0:
                msg = (_('_format_tpv, '
                         'volumename: %(volumename)s, '
                         'Return code: %(rc)lu, '
                         'Error: %(errordesc)s, '
                         'Job: %(job)s.')
                       % {'volumename': volumename,
                          'rc': rc,
                          'errordesc': errordesc,
                          'job': job})

                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            show_param_dict = {'pool-name': poolname}
            while True:
                time.sleep(wait_sec)
                rc, errordesc, job = self._exec_eternus_cli(
                    'show_tpv_progress', filename=filename, **show_param_dict)

                if rc != 0:
                    msg = (_('_format_tpv, '
                             'show volume progress error, '
                             'volumename: %(volumename)s, '
                             'Return code: %(rc)lu, '
                             'Error: %(errordesc)s, '
                             'Job: %(job)s.')
                           % {'volumename': volumename,
                              'rc': rc,
                              'errordesc': errordesc,
                              'job': job})
                    LOG.error(msg)
                    raise exception.VolumeBackendAPIException(data=msg)

                wait_sec = int(job)

                if wait_sec == 0:
                    LOG.debug('_format_tpv, format complete.')
                    break
                elif wait_cnt < 3:
                    wait_sec /= 3
                    wait_cnt += 1

                LOG.debug('_format_tpv, format wait: %s sec.', job)
        __format_tpv(self, volumename, poolname, filename)

    def initialize_connection(self, volume, connector):
        """Allow connection to connector and return connection info."""
        LOG.debug('initialize_connection, ',
                  'volume id: %(vid)s, protocol: %(prtcl)s.',
                  {'vid': volume['id'], 'prtcl': self.protocol})

        # CCM processing
        metadata = dx_utils.get_metadata(volume)
        if metadata.get(FJ_REMOTE_DEST_META, None) == FJ_REC_MIRROR:
            if volume['status'] == "in-use":
                msg = _('Live migration for backup volume is not allowed.')
                self.invalid_migration_list.append(volume['id'])
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            self._exec_ccm_script("suspend", target=metadata)

        try:
            self.conn = self._get_eternus_connection()
            vol_instance = self._find_lun(volume)
            # Check the existence of volume.
            if not vol_instance:
                # Volume isn't found.
                msg = (_('initialize_connection, '
                         'volume: %(volume)s, '
                         'volume not found.')
                       % {'volume': volume['name']})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            target_portlist, portidlist = self._get_target_port_info()
            mapdata = self._get_mapdata(vol_instance,
                                        connector, target_portlist)

            if mapdata:
                # Volume is already mapped.
                if 'target_lun' in mapdata:
                    target_lun = mapdata['target_lun']
                else:
                    target_lun = mapdata['target_luns'][0]
                msg = (_LI('initialize_connection, '
                           'volume: %(volume)s, '
                           'target_lun: %(target_lun)s, '
                           'volume is already mapped.'),
                       {'volume': volume['name'],
                        'target_lun': target_lun})
                LOG.info(msg)
            else:
                self._map_lun(vol_instance, connector, portidlist)
                mapdata = self._get_mapdata(vol_instance,
                                            connector,
                                            target_portlist)

            mapdata['target_discoverd'] = True
            mapdata['volume_id'] = volume['id']

            if self.protocol == 'fc':
                device_info = {'driver_volume_type': 'fibre_channel',
                               'data': mapdata}
            elif self.protocol == 'iSCSI':
                device_info = {'driver_volume_type': 'iscsi',
                               'data': mapdata}

        except Exception as ex:
            # When volume is set to REC Mirror, resume the session.
            if metadata.get(FJ_REMOTE_DEST_META, None) == FJ_REC_MIRROR:
                self._exec_ccm_script("resume", target=metadata)
                raise ex

        LOG.debug('initialize_connection, '
                  'device_info: %(info)s.',
                  {'info': device_info})
        return device_info

    def terminate_connection(self, volume, connector, **kwargs):
        """Disallow connection from connector."""
        LOG.debug('terminate_connection, '
                  'volume id: %(vid)s, protocol: %(prtcl)s, '
                  'options: %(options)s.',
                  {'vid': volume['id'], 'prtcl': self.protocol,
                   'options': kwargs})

        if volume['id'] in self.invalid_migration_list:
            msg = _LI('Live migration for backup volume is not allowed, '
                      'Terminate Connection Skip.')
            self.invalid_migration_list.remove(volume['id'])
            LOG.info(msg)
            self.conn = self._get_eternus_connection()
            aglist = self._find_affinity_group(connector)
            if not aglist:
                map_num = 0
            else:
                map_num = len(aglist)
            return map_num

        map_num = self.supercls.terminate_connection(volume, connector)

        # CCM processing
        metadata = dx_utils.get_metadata(volume)
        if metadata.get(FJ_REMOTE_DEST_META, None) == FJ_REC_MIRROR:
            self._exec_ccm_script("resume", target=metadata)

        LOG.debug('terminate_connection, map number: %s.', map_num)
        return map_num

    def extend_volume(self, volume, new_size):
        """Extend volume on ETERNUS."""
        LOG.debug('extend_volume, volume id: %(vid)s, '
                  'size: %(size)s, new_size: %(nsize)s.',
                  {'vid': volume['id'],
                   'size': volume['size'], 'nsize': new_size})

        self.conn = self._get_eternus_connection()

        volumename = self._get_volume_name(volume)

        # Get source volume instance.
        volume_instance = self._find_lun(volume)
        if not volume_instance:
            msg = (_('extend_volume, '
                     'source_volume: %(volumename)s, '
                     'not found.')
                   % {'volumename': volumename})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('extend_volume, volumename: %(volumename)s, '
                  'volumesize: %(volumesize)u, '
                  'source volume instance: %(volume_instance)s.',
                  {'volumename': volumename,
                   'volumesize': new_size,
                   'volume_instance': volume_instance.path})

        # Get poolname from driver configuration file.
        (pool_name, pool) = self._find_pool_from_volume(volume_instance)

        # Check the existence of pool.
        if not pool:
            msg = (_('extend_volume, '
                     'eternus_pool: %(eternus_pool)s, '
                     'not found.')
                   % {'eternus_pool': pool_name})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        # Set pooltype.
        if 'RSP' in pool['InstanceID']:
            pooltype = RAIDGROUP
        else:
            pooltype = TPPOOL

        if pooltype == RAIDGROUP:
            extend_size = str(new_size - volume['size']) + 'gb'
            param_dict = {'volume-name': volumename,
                          'rg-name': pool_name,
                          'size': extend_size}
            rc, errordesc, job = self._exec_eternus_cli('expand_volume',
                                                        **param_dict)

            if rc != 0:
                msg = (_('extend_volume, '
                         'volumename: %(volumename)s, '
                         'Return code: %(rc)lu, '
                         'Error: %(errordesc)s, '
                         'PoolType: %(pooltype)s.')
                       % {'volumename': volumename,
                          'rc': rc,
                          'errordesc': errordesc,
                          'pooltype': POOL_TYPE_dic[pooltype]})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            LOG.debug('extend_volume, '
                      'volumename: %(volumename)s, '
                      'Return code: %(rc)lu, '
                      'Error: %(errordesc)s, '
                      'Pool Name: %(eternus_pool)s, '
                      'Pool Type: %(pooltype)s.',
                      {'volumename': volumename,
                       'rc': rc,
                       'errordesc': errordesc,
                       'eternus_pool': pool_name,
                       'pooltype': POOL_TYPE_dic[pooltype]})

        else:  # pooltype is TPPOOL
            self.supercls.extend_volume(volume, new_size)

    def create_image_volume(self, volume_ref, image_id):
        """Create image volume."""
        LOG.debug('create_image_volume, volume_id: %(vid)s, '
                  'image_id: %(imid)s.',
                  {'vid': volume_ref['id'],
                   'imid': image_id})

        img_volume = {}

        img_volume['id'] = 'image-' + uuid.uuid4().hex
        img_volume['display_name'] = volume_ref['display_name']
        img_volume['size'] = volume_ref['size']
        img_volume['volume_type_id'] = volume_ref['volume_type_id']
        img_volume['volume_metadata'] = []
        storage_name = self._get_drvcfg('EternusIP')

        LOG.info(_LI('create_image_volume, '
                     'image volume id: %(img_volid)s, '
                     'storage name: %(storage_name)s.'),
                 {'img_volid': img_volume['id'],
                  'storage_name': storage_name})

        model_update = self.create_volume(img_volume)
        element_path = model_update.get('provider_location')

        img_volume['provider_location'] = element_path

        LOG.debug('create_image_volume, image volume: %s.',
                  img_volume['provider_location'])
        return img_volume

    @lockutils.synchronized('ETERNUS_DX-img-update', 'cinder-', True)
    def add_image_volume(self, volume_ref, image_id):
        """Add image volume."""
        LOG.debug('add_image_volume, image id: %s.', image_id)

        storage_name = self._get_drvcfg('EternusIP')

        with_format = dx_utils.get_extra_specs(volume_ref,
                                               key=FJ_VOL_FORMAT_KEY,
                                               default=False)
        use_format = self._get_bool(with_format)

        LOG.debug('add_image_volume, '
                  'image_id: %(image_id)s, '
                  'volume_id: %(volume_id)s, '
                  'size: %(size)s, '
                  'storage_name: %(storage_name)s.',
                  {'image_id': image_id,
                   'volume_id': volume_ref['id'],
                   'size': volume_ref['size'],
                   'storage_name': storage_name})

        self._add_image_volume_info(volume_ref['id'], volume_ref['size'],
                                    volume_ref['provider_location'],
                                    image_id, storage_name, use_format)

        LOG.debug('add_image_volume, use format: %s.', use_format)

    def _delete_image_volume(self, volume, image_id,
                             format_volume=False, delete_volume=True):
        """Delete image volume."""
        LOG.debug('_delete_image_volume, volume id: %(vol)s, '
                  'image_id: %(image_id)s.',
                  {'vol': volume['id'], 'image_id': image_id})

        @lockutils.synchronized('ETERNUS_DX-img-update', 'cinder-', True)
        def delete_image_volume_prepare(self):
            """Set delete mark to image volume management file."""
            self._update_image_volume_info(volume['id'],
                                           image_id, value=DELETE_IMGVOL)

        @lockutils.synchronized('ETERNUS_DX-img-update', 'cinder-', True)
        def delete_image_volume_update(self):
            """Remove image volume info from image volume management file."""
            self._update_image_volume_info(volume['id'], image_id, remove=True)

        delete_image_volume_prepare(self)

        if delete_volume is True:
            if format_volume is True:
                vol_instance = self._find_lun(volume)

                try:
                    assoc_class_str = 'FUJITSU_AllocatedFromStoragePool'
                    pool = (
                        self._assoc_eternus(vol_instance.path,
                                            AssocClass=assoc_class_str,
                                            ResultClass='CIM_StoragePool')[0])
                except Exception:
                    msg = (_('_delete_image_volume, '
                             'vol_instance.path: %(vol)s.')
                           % {'vol': vol_instance.path})

                    LOG.error(msg)
                    raise exception.VolumeBackendAPIException(data=msg)

                if 'RSP' in pool['InstanceID']:
                    # pooltype is RAIDGROUP
                    self._format_standard_volume(volume)
                else:
                    # pooltype is TPPOOL
                    self._format_tpv(volume, vol_instance, pool['ElementName'])

            self.delete_volume(volume)

        delete_image_volume_update(self)

    @lockutils.synchronized('ETERNUS_DX-img-update', 'cinder-', True)
    def copy_image_volume_to_volume(self, volume, image_id, src_vref=None):
        """Copy image volume to volume."""
        LOG.debug('copy_image_volume_to_volume, tgt: %(tid)s, '
                  'src: %(sid)s, image_id: %(image_id)s.',
                  {'tid': volume['id'], 'sid': src_vref,
                   'image_id': image_id})

        cloned = False
        src_volume_id = None

        # Get image volume path which meets new volume's condition.
        if not src_vref:
            image_management_file = (
                self.configuration.fujitsu_image_management_file)
            storage_name = self._get_drvcfg('EternusIP')
            src_vref = self._get_imgcfg(image_management_file,
                                        volume['size'], image_id, storage_name)

            if src_vref:
                src_volume_id = src_vref['id']

            LOG.debug('copy_image_volume_to_volume, '
                      'image_management_file_name: %(image_management_file)s, '
                      'src_volume_id: %(src_volume_id)s, '
                      'storage_name: %(storage_name)s.',
                      {'image_management_file': image_management_file,
                       'src_volume_id': src_volume_id,
                       'storage_name': storage_name})
        else:
            src_volume_id = src_vref['id']
            LOG.debug('copy_image_volume_to_volume, '
                      'src_volume_id: %(src_volume_id)s.',
                      {'src_volume_id': src_volume_id})

        # Copy image volume to volume using above image volume path.
        if src_volume_id:
            self.create_cloned_volume(volume, src_vref, CloneOnly=True)
            self._update_image_volume_info(src_volume_id, image_id)
            cloned = True

        LOG.debug('copy_image_volume_to_volume (cloned=%s).',
                  cloned)
        return cloned

    def monitor_image_volume(self):
        """Timer function for monitor image volume."""
        self.conn = self._get_eternus_connection()
        try:
            self._monitor_image_volume()
        except Exception as e:
            LOG.warning(_LW('monitor_image_volume, '
                            'undefined error was occured (%s).'),
                        e)

        timer = threading.Timer(MONITOR_IMGVOL_INTERVAL,
                                self.monitor_image_volume)
        timer.start()

    @lockutils.synchronized('ETERNUS_DX-img-monitor', 'cinder-', True)
    def _monitor_image_volume(self):
        """Monitor image volume."""
        image_management_file = (
            self.configuration.fujitsu_image_management_file)
        storage_name = self._get_drvcfg('EternusIP')
        nosession_volume_limit = (
            int(self.configuration.fujitsu_min_image_volume_per_storage))

        LOG.debug('_monitor_image_volume, '
                  'storage: %(storage_name)s, '
                  'limit: %(nosession_volume_limit)s.',
                  {'storage_name': storage_name,
                   'nosession_volume_limit': nosession_volume_limit})
        if os.path.exists(image_management_file):
            LOG.debug('monitor_image_volume, '
                      'image_management_file: %(image_management_file)s, '
                      'storage_name: %(storage_name)s, '
                      'nosession_volume_limit: %(nosession_volume_limit)s.',
                      {'image_management_file': image_management_file,
                       'storage_name': storage_name,
                       'nosession_volume_limit': nosession_volume_limit})
            doc = xml_parse(image_management_file)
            f_image = doc.getElementsByTagName('Image')

            for f_img in f_image:
                f_image_tmp = (
                    f_img.getElementsByTagName('ImageID')[0])
                f_image_id = f_image_tmp.childNodes[0].data
                f_volume = f_img.getElementsByTagName('Volume')
                nosession_volume = 0
                for f_vol in f_volume:
                    f_stor = f_vol.getElementsByTagName('StorageName')[0]
                    f_storage_name = f_stor.childNodes[0].data
                    if storage_name != f_storage_name:
                        continue

                    f_vid = f_vol.getElementsByTagName('VolumeID')[0]
                    f_volume_id = f_vid.childNodes[0].data

                    try:
                        f_vpth = f_vol.getElementsByTagName('VolumePath')[0]
                        f_volume_path = f_vpth.childNodes[0].data
                    except Exception:
                        f_volume_path = None

                    volume = {'id': f_volume_id,
                              'provider_location': f_volume_path}

                    try:
                        f_fmat = f_vol.getElementsByTagName('Format')[0]
                        f_format = f_fmat.childNodes[0].data
                        format_ = self._get_bool(f_format)
                    except Exception:
                        pass

                    try:
                        session_num = self._get_sessionnum_by_srcvol(volume)
                        if session_num == 0:
                            nosession_volume += 1
                            if nosession_volume_limit < nosession_volume:
                                LOG.info(
                                    _LI('monitor_image_volume, '
                                        'delete unused image volume which '
                                        ' is not managed by cinder and is '
                                        ' made to boost performance, '
                                        'volume id: %(volumeid)s.'),
                                    {'volumeid': f_volume_id})

                                self._delete_image_volume(
                                    volume, f_image_id, format_volume=format_)
                                continue

                        self._update_image_volume_info(f_volume_id,
                                                       f_image_id,
                                                       str(session_num))
                    except Exception as e:
                        LOG.info(_LI('monitor_image_volume, '
                                     'image volume update '
                                     'event: %(id)s (%(err)s).'),
                                 {'id': f_volume_id, 'err': e})

                        self._delete_image_volume(volume,
                                                  f_image_id,
                                                  delete_volume=False)

        LOG.debug('_monitor_image_volume, complete.')

    def _get_imgcfg(self, filename, volume_size, image_id, storage_name):
        """Read from image management file."""
        LOG.debug('_get_imgcfg input[%(filename)s][%(image_id)s][%(vsize)s].',
                  {'filename': filename, 'image_id': image_id,
                   'vsize': volume_size})

        ret = None

        if os.path.exists(filename):
            try:
                tree = tree_parse(filename)
                elem = tree.getroot()

                for child in elem:
                    f_image_id = child.findtext(".//ImageID")

                    if image_id != f_image_id:
                        continue

                    volumes = child.findall(".//Volume")
                    for volume in volumes:
                        f_session_text = volume.findtext(".//Session")
                        if f_session_text == DELETE_IMGVOL:
                            continue

                        f_session_num = int(f_session_text)
                        f_volume_size = int(volume.findtext(".//VolumeSize"))
                        f_storage_name = volume.findtext(".//StorageName")

                        try:
                            f_volume_path = volume.findtext(".//VolumePath")
                        except Exception:
                            f_volume_path = None

                        if ((f_session_num < LIMIT_SESSION_PER_VOLUME) and
                                (f_volume_size <= volume_size) and
                                (storage_name == f_storage_name)):
                            ret = {'id': volume.findtext(".//VolumeID"),
                                   'provider_location': f_volume_path}
                            break
                        else:
                            continue

                    if ret:
                        break
            except Exception:
                LOG.info(_LI('_get_imgcfg, management file is invalid.'))

        LOG.debug('_get_imgcfg output[%s].', ret)
        return ret

    def _add_image_volume_info(self, volume_id, volume_size, volume_path,
                               image_id, storage_name, use_format):
        """Add image volume information to image management file."""
        LOG.debug('_add_image_volume_info, volume_id: %(vid)s, '
                  'image_id: %(image_id)s.',
                  {'vid': volume_id, 'image_id': image_id})

        image_management_file = (
            self.configuration.fujitsu_image_management_file)

        # If file is not exist, make formatted file.
        if not os.path.exists(image_management_file):
            LOG.debug('_add_image_volume_info, create new management file.')
            doc = Document()
            root = doc.createElement('FUJITSU')
            doc.appendChild(root)

            f = codecs.open(image_management_file, 'wb', 'UTF-8')
            doc.writexml(f, '', '  ', '\n', encoding='UTF-8')
            f.close()

        # Add image volume information.
        doc = xml_parse(image_management_file)
        root = doc.documentElement
        image = doc.getElementsByTagName('Image')

        for img in image:
            f_image_id = (
                img.getElementsByTagName('ImageID')[0].childNodes[0].data)
            if image_id != f_image_id:
                continue

            volume = doc.createElement('Volume')
            img.appendChild(doc.createTextNode('\n   '))
            img.appendChild(volume)
            break
        else:
            image = doc.createElement('Image')
            root.appendChild(doc.createTextNode('\n '))
            root.appendChild(image)

            f_image_id = doc.createElement('ImageID')
            image.appendChild(f_image_id)
            f_image_id.appendChild(doc.createTextNode(image_id))

            volume = doc.createElement('Volume')
            image.appendChild(doc.createTextNode('\n   '))
            image.appendChild(volume)

        f_volume_id = doc.createElement('VolumeID')
        volume.appendChild(f_volume_id)
        f_volume_id.appendChild(doc.createTextNode(volume_id))

        f_volume_size = doc.createElement('VolumeSize')
        volume.appendChild(f_volume_size)
        f_volume_size.appendChild(doc.createTextNode(str(volume_size)))

        f_volume_path = doc.createElement('VolumePath')
        volume.appendChild(f_volume_path)
        f_volume_path.appendChild(doc.createTextNode(volume_path))

        f_storage_name = doc.createElement('StorageName')
        volume.appendChild(f_storage_name)
        f_storage_name.appendChild(doc.createTextNode(storage_name))

        f_session = doc.createElement('Session')
        volume.appendChild(f_session)
        f_session.appendChild(doc.createTextNode('0'))

        f_format = doc.createElement('Format')
        volume.appendChild(f_format)
        f_format.appendChild(doc.createTextNode(str(use_format)))

        f = codecs.open(image_management_file, 'wb', 'UTF-8')
        doc.writexml(f, encoding='UTF-8')
        f.close()
        LOG.debug('_add_image_volume_info '
                  'image_management_file: %(image_management_file)s, '
                  'image_id: %(image_id)s, '
                  'volume_id: %(volume_id)s, '
                  'volume_size: %(volume_size)s, '
                  'storage_name: %(storage_name)s.',
                  {'image_management_file': image_management_file,
                   'image_id': image_id,
                   'volume_id': volume_id,
                   'volume_size': volume_size,
                   'storage_name': storage_name})

    def _update_image_volume_info(self,
                                  volume_id,
                                  image_id, value=None, remove=False):
        """Update image volume information in image management file."""
        LOG.debug('_update_image_volume_info, volume id: %(vid)s, '
                  'image_id: %(image_id)s.',
                  {'vid': volume_id, 'image_id': image_id})

        f_image = None
        f_volume = None

        image_management_file = (
            self.configuration.fujitsu_image_management_file)
        doc = xml_parse(image_management_file)

        for f_img in doc.getElementsByTagName('Image'):
            f_image_id = (
                f_img.getElementsByTagName('ImageID')[0].childNodes[0].data)
            if f_image_id == image_id:
                f_image = f_img
                break

        if f_image:
            for f_vol in f_image.getElementsByTagName('Volume'):
                f_volid = f_vol.getElementsByTagName('VolumeID')[0]
                f_volume_id = f_volid.childNodes[0].data
                if f_volume_id == volume_id:
                    f_volume = f_vol
                    break

        if f_volume:
            if remove is False:
                f_session = f_vol.getElementsByTagName('Session')

                if not value:
                    f_session_num = (
                        str(int(f_session[0].childNodes[0].data) + 1))
                else:
                    f_session_num = value

                f_session[0].childNodes[0].data = f_session_num
                LOG.debug('_update_image_volume_info, update, '
                          'image_id: %(image_id)s, '
                          'volume_id: %(volume_id)s, '
                          'session_num: %(session_num)s.',
                          {'image_id': f_image_id,
                           'volume_id': f_volume_id,
                           'session_num': f_session_num})
            else:
                f_vol.previousSibling.data = ''
                f_img.removeChild(f_vol)
                LOG.debug('_update_image_volume_info, remove, '
                          'image_id: %(image_id)s, '
                          'volume_id: %(volume_id)s.',
                          {'image_id': f_image_id,
                           'volume_id': f_volume_id})

            f = codecs.open(image_management_file, 'wb', 'UTF-8')
            doc.writexml(f, encoding='UTF-8')
            f.close()

    def _get_sessionnum_by_srcvol(self, volume):
        """Get the number of session where specified volume is source."""
        LOG.debug('_get_sessionnum_by_srcvol, volume_id: %(vid)s.',
                  {'vid': volume['id']})

        session_info = []

        vol_instance = self._find_lun(volume)
        result_class_str = 'FUJITSU_StorageSynchronized'
        all_session_info = (
            self._reference_eternus_names(vol_instance.path,
                                          ResultClass=result_class_str))

        for session in all_session_info:
            if vol_instance.path != session['SyncedElement']:
                session_info.append(session)

        session_num = len(session_info)
        LOG.debug('_get_sessionnum_by_srcvol, '
                  ' session_num: %(session_num)s, '
                  ' session_info: %(session_info)s.',
                  {'session_num': session_num,
                   'session_info': session_info})

        return session_num

    def _set_qos(self, volume, filename=None, use_id=False):
        """Set qos using ETERNUS CLI."""
        LOG.debug('_set_qos, filename: %s.', filename)
        try:

            category = 0

            qos_specs_dict = self._get_qos_specs(volume)
            for key, value in qos_specs_dict.iteritems():
                if key in FJ_QOS_KEY_list:
                    category = self._get_qos_category_by_value(key, value)
                    break

            if 0 < category:
                volumename = self._get_volume_name(volume, use_id=use_id)
                param_dict = {'volume-name': volumename,
                              'bandwidth-limit': str(category)}
                rc, errordesc, job = self._exec_eternus_cli('set_volume_qos',
                                                            filename=filename,
                                                            **param_dict)
                if rc != 0:
                    msg = (_LW('_set_qos, '
                               'Return code: %(rc)lu, '
                               'Error: %(errordesc)s, '
                               'Job: %(job)s.')
                           % {'rc': rc,
                              'errordesc': errordesc,
                              'job': job})
                    LOG.warning(msg)
                    raise exception.VolumeBackendAPIException(data=msg)

        except Exception:
            LOG.warning(_LW('_set_qos, failed.'))

    def _get_qos_category_by_value(self, key, value):
        """Get qos category using value."""
        LOG.debug('_get_qos_category_by_value, '
                  'key: %(key)s, value: %(value)s.',
                  {'key': key, 'value': value})

        ret = 0

        # log error method
        def _get_qos_category_by_value_error():
            """Input value is invalid, log error and raise exception."""
            msg = (_LW('_get_qos_category_by_value, '
                       'Invalid value is input, '
                       'key: %(key)s, '
                       'value: %(value)s.')
                   % {'key': key,
                      'value': value})
            LOG.warning(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        if (key == "maxBWS"):
            try:
                digit = int(float(value))
            except Exception:
                _get_qos_category_by_value_error()

            if 800 <= digit:
                ret = 1
            elif 700 <= digit:
                ret = 2
            elif 600 <= digit:
                ret = 3
            elif 500 <= digit:
                ret = 4
            elif 400 <= digit:
                ret = 5
            elif 300 <= digit:
                ret = 6
            elif 200 <= digit:
                ret = 7
            elif 100 <= digit:
                ret = 8
            elif 70 <= digit:
                ret = 9
            elif 40 <= digit:
                ret = 10
            elif 25 <= digit:
                ret = 11
            elif 20 <= digit:
                ret = 12
            elif 15 <= digit:
                ret = 13
            elif 10 <= digit:
                ret = 14
            elif 0 < digit:
                ret = 15
            else:
                _get_qos_category_by_value_error()

        LOG.debug('_get_qos_category_by_value (%s).', ret)
        return ret

    def _get_target_port_info(self):
        """Return target_portid."""
        lunportlist = []
        target_portlist = []
        target_cliportidlist = []

        lunportinslist = (
            self._enum_eternus_instances('FUJITSU_LUNMappingController'))

        for lunportins in lunportinslist:
            lunportlist.append(
                self._conv_port_id(
                    lunportins['CeID'] if 'CeID' in lunportins else None,
                    lunportins['CMSlotNumber'],
                    lunportins['CASlotNumber'],
                    lunportins['PortNo']))

        if self.protocol == 'fc':
            prtcl_endpoint = 'FUJITSU_SCSIProtocolEndpoint'
            connection_type = 2
        elif self.protocol == 'iSCSI':
            prtcl_endpoint = 'FUJITSU_iSCSIProtocolEndpoint'
            connection_type = 7

        try:
            tgtportlist = self._enum_eternus_instances(prtcl_endpoint)
        except Exception:
            msg = _('_get_target_port_info, '
                    'EnumerateInstances, '
                    'cannot connect to ETERNUS.')
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        for tgtport in tgtportlist:
            cliportid = (
                self._conv_port_id(
                    tgtport['CeID'] if 'CeID' in tgtport else None,
                    tgtport['CMSlotNumber'],
                    tgtport['CASlotNumber'],
                    tgtport['PortNo']))
            if tgtport['ConnectionType'] != connection_type:
                continue
            if (tgtport['RAMode'] & 0x7B) != 0x00:
                continue
            if 'SCGroupNo' in tgtport:
                continue
            if cliportid in lunportlist:
                continue
            target_portlist.append(tgtport)
            target_cliportidlist.append(cliportid)

            LOG.debug('_get_target_port_info, '
                      'connection type: %(cont)s, '
                      'ramode: %(ramode)s.',
                      {'cont': tgtport['ConnectionType'],
                       'ramode': tgtport['RAMode']})

        LOG.debug('_get_target_port_info, '
                  'target port: %(target_portid)s.',
                  {'target_portid': target_portlist})
        LOG.debug('_get_target_port_info, '
                  'target portid: %(target_portid)s.',
                  {'target_portid': target_cliportidlist})

        if not target_portlist:
            msg = (_('_get_target_port_info, '
                     'protcol: %(protocol)s, '
                     'target_port does not found.')
                   % {'protocol': self.protocol})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        return (target_portlist, target_cliportidlist)

    @dx_utils.FJDXLockutils('connect', 'cinder-', True)
    def _map_lun(self, vol_instance, connector, portidlist=None):
        """Map volume to host."""
        volumename = vol_instance['ElementName']
        LOG.debug('_map_lun, '
                  'volumename: %(volumename)s, '
                  'connector: %(con)s.',
                  {'volumename': volumename,
                   'con': connector})

        devid_preset = set()

        volume_uid = vol_instance['Name']
        volume_lun = vol_instance['LUN']
        initiatorlist = self._find_initiator_names(connector)
        aglist = self._find_affinity_group(connector)
        configservice = self._find_eternus_service(CTRL_CONF)

        if not portidlist:
            _x, portidlist = self._get_target_port_info()

        if not configservice:
            msg = (_('_map_lun, '
                     'vol_instance.path: %(vol)s, '
                     'volumename: %(volumename)s, '
                     'volume_uid: %(uid)s, '
                     'volume_lun: %(lun)s, '
                     'initiator: %(initiator)s, '
                     'portid: %(portid)s, '
                     'aglist: %(aglist)s, '
                     'Cannot find Controller Configuration.')
                   % {'vol': vol_instance.path,
                      'volumename': [volumename],
                      'uid': [volume_uid],
                      'lun': [volume_lun],
                      'initiator': initiatorlist,
                      'portid': portidlist,
                      'aglist': aglist})

            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        LOG.debug('_map_lun, '
                  'vol_instance.path: %(vol)s, '
                  'volumename: %(volumename)s, '
                  'initiator: %(initiator)s, '
                  'portid: %(portid)s.',
                  {'vol': vol_instance.path,
                   'volumename': [volumename],
                   'initiator': initiatorlist,
                   'portid': portidlist})

        if not aglist:
            initiatorlistwk = [sss.lower() for sss in initiatorlist]
            hostnolist = []
            command = None
            hostname = None

            if 'SCGroupNo' in vol_instance:
                msg = (_('_map_lun, '
                         'vol_instance.path: %(vol)s, '
                         'volumename: %(volumename)s, '
                         'volume_uid: %(uid)s, '
                         'volume_lun: %(lun)s, '
                         'Error: %(errordesc)s.')
                       % {'vol': vol_instance.path,
                          'volumename': [volumename],
                          'uid': [volume_uid],
                          'lun': [volume_lun],
                          'errordesc': RETCODE_dic['32812']})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            # Get host name.
            if self.protocol == 'fc':
                command = 'show_host_wwn_names'
            elif self.protocol == 'iSCSI':
                command = 'show_host_iscsi_names'

            if command:
                rc, emsg, clidata = self._exec_eternus_cli(command)

                if rc != 0:
                    msg = (_('_map_lun, '
                             'Return code: %(rc)lu, '
                             'Error code: %(clidata)s, '
                             'Message: %(emsg)s.')
                           % {'rc': rc,
                              'clidata': clidata,
                              'emsg': emsg})
                    LOG.error(msg)
                    raise exception.VolumeBackendAPIException(data=msg)

                for hostdata in clidata:
                    if hostdata['Host Name'].lower() in initiatorlistwk:
                        hostnolist.append(str(hostdata['Host Num']))
                        initiatorlistwk.remove(hostdata['Host Name'].lower())

            # Create new host.
            if self.protocol == 'fc':
                command = 'create_host_wwn_name'
                hostname = 'wwn'
            elif self.protocol == 'iSCSI':
                command = 'create_host_iscsi_name'
                hostname = 'iscsi-name'

            if command:
                for initiator in initiatorlist:
                    if initiator.lower() in initiatorlistwk:
                        option = {hostname: initiator}
                        rc, emsg, clidata = self._exec_eternus_cli(command,
                                                                   **option)

                        if rc != 0:
                            msg = (_('_map_lun, '
                                     'Return code: %(rc)lu, '
                                     'Error code: %(clidata)s, '
                                     'Message: %(emsg)s.')
                                   % {'rc': rc,
                                      'clidata': clidata,
                                      'emsg': emsg})
                            LOG.error(msg)
                            raise exception.VolumeBackendAPIException(data=msg)
                        try:
                            hostnolist.append(str(int(clidata[0], 16)))
                        except Exception:
                            msg = (_('_map_lun, '
                                     'Return code: %(rc)lu, '
                                     'Error code: %(clidata)s, '
                                     'Message: %(emsg)s.')
                                   % {'rc': rc,
                                      'clidata': clidata,
                                      'emsg': emsg})
                            LOG.error(msg)
                            raise exception.VolumeBackendAPIException(data=msg)

            # Create new affinity group.
            option = {'volume-number': volume_lun,
                      'lun': '0'}
            rc, emsg, clidata = self._exec_eternus_cli('create_affinity_group',
                                                       **option)

            if rc != 0:
                msg = (_('_map_lun,'
                         'Return code: %(rc)lu, '
                         'Error code: %(clidata)s, '
                         'Message: %(emsg)s.')
                       % {'rc': rc,
                          'clidata': clidata,
                          'emsg': emsg})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            try:
                agnum = int(clidata[0], 16)
            except Exception:
                msg = (_('_map_lun, '
                         'Return code: %(rc)lu, '
                         'Error code: %(clidata)s, '
                         'Message: %(emsg)s.')
                       % {'rc': rc,
                          'clidata': clidata,
                          'emsg': emsg})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            if hostnolist:
                # Set host affinity.
                option = {'ag-number': agnum,
                          'host-number': ','.join(hostnolist),
                          'port': ','.join(portidlist)}
                rc, emsg, clidata = self._exec_eternus_cli('set_host_affinity',
                                                           **option)

                if rc != 0:
                    msg = (_('_map_lun, '
                             'Return code: %(rc)lu, '
                             'Error code: %(clidata)s, '
                             'Message: %(emsg)s.')
                           % {'rc': rc,
                              'clidata': clidata,
                              'emsg': emsg})
                    LOG.error(msg)

                    # Delete affinity group.
                    option = {'ag-number': agnum}
                    rc, emsg, clidata = (
                        self._exec_eternus_cli('delete_affinity_group',
                                               **option))

                    if rc != 0:
                        msg = (_('_map_lun, '
                                 'Return code: %(rc)lu, '
                                 'Error code: %(clidata)s, '
                                 'Message: %(emsg)s.')
                               % {'rc': rc,
                                  'clidata': clidata,
                                  'emsg': emsg})
                        LOG.error(msg)

                    raise exception.VolumeBackendAPIException(data=msg)
        else:
            # Add lun to affinity group.
            for ag in aglist:
                LOG.debug('_map_lun, '
                          'ag: %(ag)s, '
                          'lun_name: %(volume_uid)s.',
                          {'ag': ag,
                           'volume_uid': volume_uid})

                devid_pre = ag['DeviceID'][:8]

                if devid_pre not in devid_preset:
                    rc, errordesc, job = self._exec_eternus_service(
                        'ExposePaths',
                        configservice, LUNames=[volume_uid],
                        DeviceAccesses=[pywbem.Uint16(2)],
                        ProtocolControllers=[ag])

                    LOG.debug('_map_lun, '
                              'Error: %(errordesc)s, '
                              'Return code: %(rc)lu, '
                              'Add lun affinitygroup.',
                              {'errordesc': errordesc,
                               'rc': rc})

                    if rc != 0:
                        msg = (_LE('_map_lun, '
                                   'lun_name: %(volume_uid)s, '
                                   'Initiator: %(initiator)s, '
                                   'portid: %(portid)s, '
                                   'Return code: %(rc)lu, '
                                   'Error: %(errordesc)s.'),
                               {'volume_uid': [volume_uid],
                                'initiator': initiatorlist,
                                'portid': portidlist,
                                'rc': rc,
                                'errordesc': errordesc})
                        LOG.error(msg)

                    devid_preset.add(devid_pre)
        LOG.debug('_map_lun, '
                  'volumename: %(volumename)s.',
                  {'volumename': volumename})

    def _check_user(self):
        """Check whether user's role is accessible to ETERNUS and Software."""
        ret = True
        rc, errordesc, job = self._exec_eternus_cli('check_user_role')
        if rc != 0:
            msg = (_('_check_user, '
                     'Return code: %(rc)lu, '
                     'Error: %(errordesc)s, '
                     'Job: %(job)s.')
                   % {'rc': rc,
                      'errordesc': errordesc,
                      'job': job})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        if job != 'Software':
            msg = (_('_check_user, '
                     'Specified user(%(user)s) does not have '
                     'Software role: %(role)s.')
                   % {'user': self._get_drvcfg('EternusUser'),
                      'role': job})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        return ret

    @dx_utils.FJDXLockutils('SMIS-exec', 'cinder-', True)
    def _exec_eternus_cli(self, command, retry=20, retry_interval=5,
                          retry_code=[32787], filename=None, **param_dict):
        """Execute ETERNUS CLI."""
        LOG.debug('_exec_eternus_cli, '
                  'command: %(a)s, '
                  'filename: %(f)s, '
                  'parameters: %(b)s.',
                  {'a': command,
                   'f': filename,
                   'b': param_dict})

        out = None
        _err = None
        result = None
        rc = None
        retdata = None
        errordesc = None
        user = self._get_drvcfg('EternusUser', filename)
        storage_ip = self._get_drvcfg('EternusIP', filename)

        for retry_num in range(retry):
            # Execute ETERNUS CLI and get return value.
            try:
                if param_dict:
                    (out, _err) = utils.execute('eternus_dx_cli', '-u', user,
                                                '-l', storage_ip,
                                                '-c', command,
                                                '-o', str(param_dict),
                                                run_as_root=True)
                else:
                    (out, _err) = utils.execute('eternus_dx_cli', '-u', user,
                                                '-l', storage_ip,
                                                '-c', command,
                                                run_as_root=True)

                out_dict = eval(out)
                result = out_dict.get('result')
                rc_str = out_dict.get('rc')
                retdata = out_dict.get('message')
            except Exception as ex:
                msg = (_('_exec_eternus_cli, '
                         'stdout: %(out)s, '
                         'stderr: %(err)s, '
                         'unexpected error: %(ex)s.')
                       % {'out': out,
                          'err': _err,
                          'ex': ex})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            # Check ssh result.
            if result == 255:
                LOG.info(_LI('_exec_eternus_cli, retry, '
                             'command: %(command)s, '
                             'option: %(option)s, '
                             'ip: %(ip)s, '
                             'SSH Result: %(result)s, '
                             'retdata: %(retdata)s, '
                             'TryNum: %(rn)s.'),
                         {'command': command,
                          'option': param_dict,
                          'ip': storage_ip,
                          'result': result,
                          'retdata': retdata,
                          'rn': (retry_num + 1)})
                time.sleep(retry_interval)
                continue
            elif result != 0:
                msg = (_('_exec_eternus_cli, '
                         'unexpected error, '
                         'command: %(command)s, '
                         'option: %(option)s, '
                         'ip: %(ip)s, '
                         'result: %(result)s, '
                         'retdata: %(retdata)s.')
                       % {'command': command,
                          'option': param_dict,
                          'ip': storage_ip,
                          'result': result,
                          'retdata': retdata})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            # Check CLI return code.
            if rc_str.isdigit():
                # SMI-S style return code
                rc = int(rc_str)

                try:
                    errordesc = RETCODE_dic[str(rc)]
                except Exception:
                    errordesc = 'Undefined Error!!'

                if rc in retry_code:
                    LOG.info(_LI('_exec_eternus_cli, retry, '
                                 'ip: %(ip)s, '
                                 'RetryCode: %(rc)s, '
                                 'TryNum: %(rn)s.'),
                             {'ip': storage_ip,
                              'rc': rc,
                              'rn': (retry_num + 1)})
                    time.sleep(retry_interval)
                    continue

                break
            else:
                # CLI style return code
                LOG.warning(_LW('_exec_eternus_cli, '
                                'WARNING!! '
                                'ip: %(ip)s, '
                                'ReturnCode: %(rc_str)s, '
                                'ReturnData: %(retdata)s.'),
                            {'ip': storage_ip,
                             'rc_str': rc_str,
                             'retdata': retdata})

                errordesc = retdata
                retdata = rc_str
                rc = 4  # Failed
                break
        else:
            if 0 < result:
                msg = (_('_exec_eternus_cli, '
                         'cannot connect to ETERNUS. '
                         'SSH Result: %(result)s, '
                         'retdata: %(retdata)s.')
                       % {'result': result,
                          'retdata': retdata})

                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            else:
                LOG.warning(_LW('_exec_eternus_cli, Retry was exceeded.'))

        ret = (rc, errordesc, retdata)

        LOG.debug('_exec_eternus_cli, '
                  'command: %(a)s, '
                  'parameters: %(b)s, '
                  'ip: %(ip)s, '
                  'Return code: %(rc)s, '
                  'Error: %(errordesc)s.',
                  {'a': command,
                   'b': param_dict,
                   'ip': storage_ip,
                   'rc': rc,
                   'errordesc': errordesc})
        return ret

    def _get_qos_specs(self, volume):
        """Get qos specs information from volume information."""
        LOG.debug('_get_qos_specs, volume id: %(volid)s.',
                  {'volid': volume['id']})

        qos_specs_dict = {}
        qos_specs_id = None

        volume_type_id = volume.get('volume_type_id')

        if volume_type_id:
            ctxt = context.get_admin_context()
            volume_type = volume_types.get_volume_type(ctxt, volume_type_id)
            qos_specs_id = volume_type.get('qos_specs_id')

        if qos_specs_id:
            qos_specs_dict = qos_specs.get_qos_specs(ctxt,
                                                     qos_specs_id)['specs']

        LOG.debug('_get_qos_specs, qos_specs_dict: %(qos_dict)s.',
                  {'qos_dict': qos_specs_dict})
        return qos_specs_dict

    def _get_bool(self, value):
        """Get True/False from String 'True'/'False'."""
        LOG.debug('_get_bool, value: %(value)s.',
                  {'value': value})

        ret = False

        if str(value).lower() == "true":
            ret = True

        return ret

    def _conv_port_id(self, ce_id, cm_slot_num, ca_slot_num, ca_port_num):
        """Convert port info to standard string."""
        if ce_id is None:
            port_id = '%x%x%x' % (cm_slot_num, ca_slot_num, ca_port_num)
        else:
            port_id = '%x%x%x%x' % (ce_id,
                                    cm_slot_num,
                                    ca_slot_num,
                                    ca_port_num)

        return port_id

    def _delete_volume_base(self, volume, with_format, remote=False):
        LOG.debug('_delete_volume_base, remote: %(remote)s.',
                  {'remote': remote})

        if not remote:
            conn = self.conn
            vol_instance = self._find_lun(volume)
        else:
            conn = self.remote_conn
            vol_instance = self._find_replica_lun(volume)

        if not vol_instance:
            return True

        if self._get_bool(with_format):
            poolname, pool = self._find_pool_from_volume(vol_instance,
                                                         remote=remote)
            if not pool:
                msg = (_('_delete_volume_base, '
                         'vol_instance: %(vol_instance)s, '
                         'pool not found.')
                       % {'vol_instance': vol_instance.path})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            if 'RSP' in pool['InstanceID']:  # pooltype == RAIDGROUP
                self._format_standard_volume(volume, remote=remote)
            else:  # pooltype == TPPOOL
                self._format_tpv(volume, vol_instance, poolname, remote=remote)

        if vol_instance:
            self._delete_volume(volume, conn=conn, remote=remote)

        return True

    def manage_existing(self, volume, existing_ref):
        """Bring an existing storage object under Cinder management.

        existing_ref can contain source-id or source-name.
        source-id: lun uuid
        source-name: lun name
        """
        model_update = self.supercls.manage_existing(volume, existing_ref)
        volume['provider_location'] = model_update['provider_location']
        self._set_qos(volume)
        return model_update
