# Copyright (c) 2016 FUJITSU LIMITED
# Copyright (c) 2012 EMC Corporation.
# Copyright (c) 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""
iSCSI Cinder Volume driver for Fujitsu ETERNUS DX S3 series.

This driver is extension version.
"""

# $Date:: 2016-02-17 11:56:58 +0800#$
# $Revision: 309 $

import threading

from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_log import log as logging

from cinder import context
from cinder import exception
from cinder import utils
from cinder.i18n import _, _LI, _LW
from cinder.volume.drivers.fujitsu import eternus_dx_iscsi
from cinder.volume.drivers.fujitsu import eternus_dx_utils as dx_utils
from cinder.volume.drivers.fujitsu.extension import eternus_dx_common
from cinder.volume.drivers.fujitsu.extension import image_utils

LOG = logging.getLogger(__name__)

FJ_SRC_VOL_ID = "FJ_Source_Volume_ID"
FJ_REMOTE_DEST_META = "FJ_Remote_Copy_Destination"
FJ_ETERNUS_DX_OPT_list = [
    cfg.BoolOpt('use_fujitsu_image_volume',
                default=True,
                help='Use image volume when '
                'copying image to volume.'),
]


class FJDXISCSIDriver(eternus_dx_iscsi.FJDXISCSIDriver):
    """ETERNUS iSCSI Cinder Volume Driver for Fujitsu ETERNUS DX S3 series."""

    def __init__(self, *args, **kwargs):
        super(FJDXISCSIDriver, self).__init__(*args, **kwargs)
        self.configuration.append_config_values(FJ_ETERNUS_DX_OPT_list)
        self.common = (
            eternus_dx_common.FJDXCommonExt(
                'iSCSI', configuration=self.configuration))
        self.first_loop = True
        self.VERSION = self.common.VERSION

    def create_volume(self, volume):
        """Create volume."""
        LOG.info(_LI('create_volume, '
                     'volume id: %s, Enter method.'), volume['id'])

        # Get volume metadata.
        d_metadata = dx_utils.get_metadata(volume)

        # Choice: standard create volume or CCM create volume.
        if FJ_SRC_VOL_ID not in d_metadata:
            # Standard create volume.
            model_update = self.common.create_volume(volume)
        else:
            # CCM create volume.
            # Create cloned volume for backup solution.
            # Get source and target volume information.
            change_metadata = {}
            ctxt = context.get_admin_context()
            volume_size = volume['size']
            src_volid = d_metadata[FJ_SRC_VOL_ID]
            if src_volid is None:
                msg = _('create_volume, Source volume id is None.')
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            # Modify metadata value Mirror to 'mirror'.
            if FJ_REMOTE_DEST_META in d_metadata:
                change_metadata[FJ_REMOTE_DEST_META] = (
                    (d_metadata[FJ_REMOTE_DEST_META]).lower())

            # Get info of source volume.
            src_vref = self.db.volume_get(ctxt, src_volid)

            # Check target volume information with reference to source volume.
            src_vref_size = src_vref['size']
            src_vref_status = src_vref['status']

            if (('error' in src_vref_status) or
                    ('deleting' in src_vref_status)):
                msg = (_('Invalid volume status: %(status)s.')
                       % {'status': src_vref_status})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)
            elif volume_size < src_vref_size:
                msg = (_('Volume size %(volume_size)sGB '
                         'cannot be smaller than '
                         'original volume size %(src_vref_size)sGB. '
                         'They must be >= original volume size.')
                       % {'volume_size': volume_size,
                          'src_vref_size': src_vref_size})
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            LOG.info(_LI('create_volume, Call create cloned volume.'))
            model_update = self.common.create_cloned_volume(
                volume, src_vref, use_service_name=True)

            LOG.info(_LI('create_volume, Return create cloned volume.'))

            # post processing
            try:
                if src_vref.bootable:
                    self.db.volume_update(ctxt,
                                          volume['id'],
                                          {'bootable': True})
                    self.db.volume_glance_metadata_copy_from_volume_to_volume(
                        ctxt, src_volid, volume['id'])

                self.db.volume_update(ctxt,
                                      volume['id'],
                                      {'source_volid': src_volid})
            except Exception as ex:
                msg = (_('create volume, Failed updating volume metadata, '
                         'reason: %(reason)s, volume_id: %(volume_id)s, '
                         'src_volid: %(src_volid)s.')
                       % {'reason': ex,
                          'volume_id': volume['id'],
                          'src_volid': src_volid})
                raise exception.MetadataCopyFailure(reason=msg)

            model_update['metadata'].update(change_metadata)

        LOG.info(_LI('create_volume, '
                     'info: %s, Exit method.'), model_update['metadata'])
        return model_update

    def create_volume_from_snapshot(self, volume, snapshot):
        """Create a volume from a snapshot."""
        LOG.info(_LI('create_volume_from_snapshot, '
                     'volume id: %(vid)s, snap id: %(sid)s, Enter method.'),
                 {'vid': volume['id'], 'sid': snapshot['id']})

        model_update = (
            self.common.create_volume_from_snapshot(volume, snapshot))

        LOG.info(_LI('create_volume_from_snapshot, '
                     'info: %s, Exit method.'), model_update['metadata'])
        return model_update

    def create_cloned_volume(self, volume, src_vref):
        """Create cloned volume."""
        LOG.info(_LI('create_cloned_volume, '
                     'target volume id: %(tid)s, '
                     'source volume id: %(sid)s, Enter method.'),
                 {'tid': volume['id'], 'sid': src_vref['id']})

        model_update = self.common.create_cloned_volume(volume, src_vref)

        LOG.info(_LI('create_cloned_volume, '
                     'info: %s, Exit method.'), model_update['metadata'])
        return model_update

    def delete_volume(self, volume):
        """Delete volume on ETERNUS."""
        LOG.info(_LI('delete_volume, '
                     'volume id: %s, Enter method.'), volume['id'])

        vol_exist = self.common.delete_volume(volume)

        LOG.info(_LI('delete_volume, '
                     'delete: %s, Exit method.'), vol_exist)

    def initialize_connection(self, volume, connector):
        """Allow connection to connector and return connection info."""
        LOG.info(_LI('initialize_connection, volume id: %(vid)s, '
                     'initiator: %(initiator)s, Enter method.'),
                 {'vid': volume['id'], 'initiator': connector['initiator']})

        info = self.common.initialize_connection(volume, connector)

        LOG.info(_LI('initialize_connection, '
                     'info: %s, Exit method.'), info)
        return info

    def terminate_connection(self, volume, connector, **kwargs):
        """Disallow connection from connector."""
        LOG.info(_LI('terminate_connection, volume id: %(vid)s, '
                     'initiator: %(initiator)s, Enter method.'),
                 {'vid': volume['id'], 'initiator': connector['initiator']})

        map_num = self.common.terminate_connection(volume, connector)

        LOG.info(_LI('terminate_connection, '
                     'map_num: %s, Exit method.'), map_num)

    def get_volume_stats(self, refresh=False):
        """Get volume stats."""
        LOG.debug('get_volume_stats, refresh: %s, Enter method.', refresh)

        pool_name = None
        if refresh is True:
            data, pool_name = self.common.update_volume_stats()
            backend_name = self.configuration.safe_get('volume_backend_name')
            data['volume_backend_name'] = backend_name or 'FJDXISCSIDriver'
            data['storage_protocol'] = 'iSCSI'
            self._stats = data

        if self.first_loop is True:
            self.first_loop = False
            monitor_thread = (
                threading.Thread(target=self.common.monitor_image_volume))
            monitor_thread.start()

        LOG.debug('get_volume_stats, '
                  'pool name: %s, Exit method.', pool_name)
        return self._stats

    def extend_volume(self, volume, new_size):
        """Extend volume."""
        LOG.info(_LI('extend_volume, '
                     'volume id: %s, Enter method.'), volume['id'])

        used_pool_name = self.common.extend_volume(volume, new_size)

        LOG.info(_LI('extend_volume, '
                     'used pool name: %s, Exit method.'), used_pool_name)

    @lockutils.synchronized('ETERNUS_DX-img', 'cinder-', True)
    def copy_image_to_volume(self, context, volume, image_service, image_id):
        """Fetch the image from image_service and write it to the volume."""
        LOG.info(_LI('copy_image_to_volume, '
                     'use_image_volume: %s, Enter method.'),
                 self.configuration.use_fujitsu_image_volume)

        cloned = False

        setup = self._setup_image_volume(volume)

        use_multipath = self.configuration.use_multipath_for_image_xfer
        enforce_multipath = (
            self.configuration.enforce_multipath_for_image_xfer)
        properties = utils.brick_get_connector_properties(
            use_multipath, enforce_multipath)

        if self.configuration.use_fujitsu_image_volume is True:

            try:  # Creating image volume using dd command.
                if setup is False:
                    cloned = self._copy_image_volume_to_volume(volume,
                                                               image_id)

                if cloned is True:
                    LOG.info(_LI('copy_image_to_volume, '
                                 'create volume using OPC.'))
                else:
                    image_volume = self._create_image_volume(volume, image_id)

                    try:
                        self._write_image_to_volume(context,
                                                    properties,
                                                    image_volume,
                                                    image_service,
                                                    image_id)
                        self._add_image_volume(image_volume, image_id)
                    except Exception as ex:
                        LOG.info(_LI('%s'), ex)
                        self.delete_volume(image_volume)
                        raise ex

                    if setup is False:
                        cloned = self._copy_image_volume_to_volume(
                            volume, image_id, src_vref=image_volume)
                        LOG.info(_LI('copy_image_to_volume, '
                                     'create volume by writing image, '
                                     'result: %s.'),
                                 cloned)
            except Exception as ex:
                LOG.warning(_LW('%s'), ex)
                LOG.warning(_LW('copy_image_to_volume, create volume '
                                'without creating image volume.'))
                if setup is False:
                    self._write_image_to_volume(context, properties,
                                                volume, image_service,
                                                image_id)
                else:
                    raise ex
        elif setup is True:  # use_fujitsu_image_volume=False & setup=True
            msg = _('copy_image_to_volume, this backend specify '
                    'that use_fujitsu_image_volume is False.')
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)
        else:  # if not use_fujitsu_image_volume or setup
            properties = utils.brick_get_connector_properties(
                use_multipath, enforce_multipath)
            self._write_image_to_volume(context, properties,
                                        volume, image_service, image_id)

    def _write_image_to_volume(self, context, properties,
                               volume, image_service, image_id):
        """Fetch the image and write to the volume using dd command."""
        LOG.debug('_write_image_to_volume, image id: %s, Enter method.',
                  image_id)

        attach_info, volume = self._attach_volume(context, volume, properties)

        try:
            image_utils.fetch_to_raw(context,
                                     image_service,
                                     image_id,
                                     attach_info['device']['path'],
                                     self.configuration.volume_dd_blocksize,
                                     size=volume['size'])
        finally:
            self._detach_volume(context, attach_info, volume, properties)

    def _setup_image_volume(self, volume):
        """Check whether setup mode or not, and prepare for setup mode."""
        LOG.debug('_setup_image_volume, volume id: %s, Enter method.',
                  volume['id'])

        setup = False

        meta = volume.get('volume_metadata')
        for data in meta:
            if data['key'] == 'fujitsu_image_setup':
                if data['value'] == 'True':
                    setup = True
                    volume_type_id = volume['volume_type_id']
                    volume['volume_type_id'] = None
                    self.common.delete_volume(volume)
                    volume['volume_type_id'] = volume_type_id
                    LOG.info(_LI('_setup_image_volume, setup: %s.'), setup)
                    break

        LOG.debug('_setup_image_volume (setup: %s), Exit method.', setup)
        return setup

    def _create_image_volume(self, volume_ref, image_id):
        """Create blank volume for image volume."""
        LOG.debug('_create_image_volume, image id: %s, Enter method.',
                  image_id)

        image_volume = self.common.create_image_volume(volume_ref, image_id)

        LOG.debug('_create_image_volume, image volume: %s, Exit method.',
                  image_volume['provider_location'])
        return image_volume

    def _add_image_volume(self, volume_ref, image_id):
        """Create blank volume for image volume."""
        LOG.debug('_add_image_volume, image id: %s, Enter method.',
                  image_id)

        self.common.add_image_volume(volume_ref, image_id)

    def _copy_image_volume_to_volume(self, volume, image_id, src_vref=None):
        """Copy image volume to volume."""
        LOG.debug('_copy_image_volume_to_volume, image id: %s, Enter method.',
                  image_id)

        cloned = self.common.copy_image_volume_to_volume(volume,
                                                         image_id,
                                                         src_vref)

        LOG.debug('_copy_image_volume_to_volume (cloned: %s), Exit method.',
                  cloned)
        return cloned

    def manage_existing(self, volume, existing_ref):
        """Bring an existing volume under Cinder management."""
        LOG.info(_LI('manage_existing, volume id: %(vid)s, '
                     'id or name: %(ref)s, Enter method.'),
                 {'vid': volume['id'], 'ref': existing_ref})

        info = self.common.manage_existing(volume, existing_ref)

        LOG.info(_LI('manage_existing, '
                     'info: %s, Exit method.'), info['metadata'])
        return info
